(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.
Require Import Arith.

Require Import replace.

Require Import predicate.
Require Import closure.
Require Import maps.
Require Import carrier.
Require Import finite.

Require Import enum.
Require Import enumerations.

Require Import formula.

Require Import labels.
Require Import words.
Require Import constraints.
Require Import pme.

Definition stm : Type := (bool*(formula*word label))%type.
  
Definition enum_stm : enumeration stm.
Proof.
  apply enum_prod.
  apply enum_bool.
  apply enum_prod.
  apply enum_formula.
  apply enum_word.
  apply enum_from_idx.
  apply idx_label.
Qed.

Definition stm_eq_dec : forall x y : stm, { x = y } + { x <> y }.
Proof. apply enumeration_dec, enum_stm. Qed. 

Delimit Scope Stm_scope with stm.

Notation " 'T' f # m " := ((true,(f,m))) (at level 33) : Stm_scope.
Notation " 'F' f # m " := ((false,(f,m))) (at level 33) : Stm_scope.

Open Scope Stm_scope.

Definition finite_pair {X Y} c := match c with (x,y) => finite (x : X->Prop) /\ finite (y:Y->Prop) end.

Proposition finite_pair_auto X Y (x : X -> Prop) (y : Y -> Prop) : finite x -> finite y -> finite_pair (x,y).
Proof. split; auto. Qed.

Hint Extern 2 ( finite_pair (_,_) ) => split.

Proposition finite_pair_id X Y (x1 x2 : X -> Prop) (y1 y2 : Y->Prop) : x1 ~ x2 -> y1 ~ y2 -> finite_pair (x1,y1) -> finite_pair (x2,y2).
Proof. intros ? ? (H1,H2). split; [ apply finite_id with (2 := H1) | apply finite_id with (2 := H2) ]; auto. Qed. 

Hint Resolve finite_pair_id.

Definition cst := constraint label.

Definition cst_eq_dec : forall x y : cst, { x = y } + { x <> y }.
Proof. apply constraint_eq_dec, enum_from_idx, idx_label. Qed.

Hint Resolve cst_eq_dec.

Definition CSS_type := ((stm -> Prop)*(cst -> Prop))%type.

Definition CSS (p : CSS_type) := match p with (X,C) => forall s f m, X (s,(f,m)) -> pme C (m~~m) end.
Definition finite_CSS p := CSS p /\ finite_pair p.

Section CSS_rels.

  Variables p q : CSS_type.

  Definition CSS_id     := match (p,q) with ((X1,C1),(X2,C2)) => X1 ~ X2 /\ C1 ~ C2 end.
  Definition CSS_leq    := match (p,q) with ((X1,C1),(X2,C2)) => X1 inc X2 /\ C1 inc C2 end.
  Definition CSS_equiv  := match (p,q) with ((X1,C1),(X2,C2)) => X1 ~ X2 /\ pme C1 ~ pme C2 end.
  Definition CSS_lequiv := match (p,q) with ((X1,C1),(X2,C2)) => X1 inc X2 /\ pme C1 inc pme C2 end.
  Definition CSS_inc    := match (p,q) with ((X1,C1),(X2,C2)) => X1 ~ X2 /\ pme C1 inc pme C2 /\ alphabet C1 ~ alphabet C2 end.

End CSS_rels.

Hint Extern 2 ( finite_CSS (_,_) ) => split.
Hint Extern 2 ( CSS_id (_,_) (_,_) ) => split.
Hint Extern 2 ( CSS_leq (_,_) (_,_) ) => split.
Hint Extern 2 ( CSS_equiv (_,_) (_,_) ) => split.
Hint Extern 2 ( CSS_lequiv (_,_) (_,_) ) => split.

Proposition CSS_id_refl : forall p, CSS_id p p.                     Proof. intros []; auto. Qed.
Proposition CSS_leq_refl : forall p, CSS_leq p p.                   Proof. intros []; auto. Qed.
Proposition CSS_equiv_refl : forall p, CSS_equiv p p.               Proof. intros []; auto. Qed.
Proposition CSS_lequiv_refl : forall p, CSS_lequiv p p.             Proof. intros []; auto. Qed.

Hint Immediate CSS_id_refl CSS_leq_refl CSS_equiv_refl CSS_lequiv_refl.

Proposition CSS_id_sym : forall p q, CSS_id p q -> CSS_id q p.             Proof. intros [] [] []; auto. Qed.
Proposition CSS_equiv_sym : forall p q, CSS_equiv p q -> CSS_equiv q p.    Proof. intros [] [] []; auto. Qed.

Hint Immediate CSS_id_sym CSS_equiv_sym.

Section trans.

  Ltac solve H := 
    let H3 := fresh in
    let H4 := fresh in
    intros [] [] [] (?,?) (H3,H4); 
    split; [ apply H with (2 := H3) | apply H with (2 := H4) ]; auto.

  Ltac solve_inc := solve inc_transitive.
  Ltac solve_id := solve id_transitive.

  Proposition CSS_id_trans : forall p q r, CSS_id p q -> CSS_id q r -> CSS_id p r.
  Proof. solve_id. Qed.

  Proposition CSS_leq_trans : forall p q r, CSS_leq p q -> CSS_leq q r -> CSS_leq p r.
  Proof. solve_inc. Qed.

  Proposition CSS_equiv_trans : forall p q r, CSS_equiv p q -> CSS_equiv q r -> CSS_equiv p r.
  Proof. solve_id. Qed.

  Proposition CSS_lequiv_trans : forall p q r, CSS_lequiv p q -> CSS_lequiv q r -> CSS_lequiv p r.
  Proof. solve_inc. Qed.

End trans.

Hint Immediate CSS_id_trans CSS_leq_trans CSS_equiv_trans CSS_lequiv_trans.

Proposition CSS_id_id : forall css1 css2, CSS_id css1 css2 -> CSS css1 -> CSS css2.
Proof. 
  intros (X1,C1) (X2,C2) (HX,HC) H s f m H'.
  apply pme_mono with (R := C1).
  apply HC.
  apply (H s f m), HX; auto.
Qed.

Proposition CSS_equiv_id : forall css1 css2, CSS_equiv css1 css2 -> CSS css1 -> CSS css2.
Proof. intros (X1,C1) (X2,C3) (HX,HC) H s f m H'; apply (proj1 HC), (H s f m), (proj2 HX), H'. Qed.

Proposition fCSS_id : forall css1 css2, CSS_id css1 css2 -> finite_CSS css1 -> finite_CSS css2.
Proof. 
  intros (X1,C1) (X2,C2) H [ ? [ ? ? ] ]; split. 
  apply CSS_id_id with (css1 := (X1,C1)); auto.
  destruct H; split; 
  [ apply finite_id with (2 := H1) 
  | apply finite_id with (2 := H2) ]; auto.
Qed.

Proposition CSS_id_imp_leq : forall p q, CSS_id p q -> CSS_leq p q.
Proof. intros [] [] [ [] [] ]; auto. Qed.

Proposition CSS_id_imp_equiv : forall p q, CSS_id p q -> CSS_equiv p q.
Proof. intros [] [] [ ]; auto. Qed.

Hint Resolve CSS_id_imp_leq CSS_id_imp_equiv.

Proposition CSS_leq_antisym : forall p q, CSS_id p q <-> CSS_leq p q /\ CSS_leq q p.
Proof. 
  intros [] []; split. 
  intros []; split; auto.
  intros [ [] [] ]; split; auto.
Qed.

Proposition CSS_lequiv_antisym : forall p q, CSS_equiv p q <-> CSS_lequiv p q /\ CSS_lequiv q p. 
Proof. intros [] []; split; intros [ [] [] ]; split; auto; split; auto. Qed.

Hint Resolve CSS_leq_antisym CSS_lequiv_antisym.

Definition CSS_leqf p1 p2 := CSS_leq p1 p2 /\ finite_CSS p1.

Hint Extern 2 ( CSS_leqf _ _ ) => split.

Proposition CSS_leqf_refl : forall p, finite_CSS p -> CSS_leqf p p.
Proof. intros []; auto. Qed.

Proposition CSS_leqf_tran : forall p q r, CSS_leqf p q -> CSS_leqf q r -> CSS_leqf p r.
Proof. intros ? ? ? [] [ H ? ]; split; auto; apply CSS_leq_trans with (2 := H); auto. Qed.

Proposition CSS_dec_l X X' C : X inc X' -> CSS (X',C) -> CSS (X,C).
Proof. intros H1 H2 ? ? ? H3; apply H2 with (1 := H1 _ H3). Qed.

Proposition CSS_inc_empty_r X C : CSS (X, C cup empty) -> CSS (X,C).
Proof. apply CSS_equiv_id; auto. Qed.

Proposition fCSS_dec_l X X' C : X inc X' -> finite X -> finite_CSS (X',C) -> finite_CSS (X,C).
Proof. intros ? ? (?,(?,?)); split; auto; apply CSS_dec_l with (X':=X'); auto. Qed.

Hint Resolve CSS_dec_l CSS_inc_empty_r fCSS_dec_l.

Proposition CSS_cup_l X1 X2 C : CSS (X1,C) -> CSS (X2,C) -> CSS (X1 cup X2,C).
Proof. intros H1 H2 ? ? ? ?; pred_solve; [ apply H1 with (1 := H) | apply H2 with (1 := H) ]. Qed.

Proposition CSS_sg_l s f m C : pme C (m~~m) -> CSS (sg (s,(f,m)), C).
Proof. intros H ? ? ? H'; pred_solve; replace_with H; injection H'; auto. Qed.

Hint Resolve CSS_cup_l CSS_sg_l.

Proposition fCSS_sg_l s f m C : finite C -> pme C (m~~m) -> finite_CSS (sg (s,(f,m)), C).
Proof. intros; auto. Qed.

Hint Resolve fCSS_sg_l.

Proposition CSS_empty_l C : CSS (empty, C).
Proof. intros ? ? ? ?; pred_solve. Qed.

Hint Immediate CSS_empty_l.

Proposition CSS_inc_empty_l X C : CSS (X,C) -> CSS (X cup empty,C).
Proof. intros; auto. Qed. 

Hint Resolve CSS_inc_empty_l.

Proposition fCSS_inc_empty_l X C : finite_CSS (X,C) -> finite_CSS (X cup empty,C).
Proof. intros (?,(?,?)); auto. Qed. 

Hint Resolve fCSS_inc_empty_l.

Section CSS_compact.
  
  Let CSS_compact_rec l1 : forall C, CSS (carrier l1, C) -> exists Cf, Cf inc C /\ finite Cf /\ CSS (carrier l1, Cf).
  Proof.
    induction l1 as [ | (s,(f,m)) l1 ]; intros C H.
    exists empty; auto. 
    
    assert (pme C (m~~m)) as H'.
      apply (H s f m); auto.
    apply pme_cpct in H'.
    destruct H' as ( C0 & H0 & H1 & H2 ).
    destruct (IHl1 C) as ( C1 & H3 & H4 & H5 ).
      intros s' f' m' H'. 
      apply (H s' f' m'); auto.
    exists (C0 cup C1).
    split; auto.
    split; auto.
    intros s' f' m' H'.
    destruct H' as [ H' | H' ].
    injection H'; intros H'' _ _; rewrite <- H''.
    apply pme_mono with (R := C0); auto.
    apply pme_mono with (R := C1); auto.
    apply (H5 s' f' m'); auto.
  Qed.
  
  Proposition CSS_compact Xf C : CSS (Xf, C) -> finite Xf -> exists Cf, Cf inc C /\ finite_CSS (Xf, Cf).
  Proof.
    intros H H'. 
    generalize H'; intros [ l1 [ Hl1 Hl2 ] ].
    apply CSS_dec_l with (1 := Hl1) in H.
    apply CSS_compact_rec in H.
    destruct H as ( Cf & H0 & H1 & H2 ).
    exists Cf.
    split; auto.
    split; auto.
    apply CSS_dec_l with (2 := H2); auto.
  Qed.

End CSS_compact.

Proposition CSS_inc_stm_l X C s f m : CSS (X,C) -> pme C (m ~~ m) -> CSS (X cup sg (s,(f,m)), C).
Proof. auto. Qed.

Proposition CSS_cst_cup X C s f m : CSS (X cup sg (s,(f,m)),C) -> pme C (m~~m).
Proof. intros H; apply (H s f m); pred_solve. Qed.

Proposition CSS_cst_cup_cup_l X C s f m s' f' m' : CSS (X cup (sg (s,(f,m)) cup sg (s',(f',m'))),C) -> pme C (m~~m).
Proof. intros H; apply (H s f m); pred_solve. Qed.

Proposition CSS_cst_cup_cup_r X C s f m s' f' m' : CSS (X cup (sg (s,(f,m)) cup sg (s',(f',m'))),C) -> pme C (m'~~m').
Proof. intros H; apply (H s' f' m'); pred_solve. Qed.

Hint Resolve CSS_inc_stm_l CSS_cst_cup CSS_cst_cup_cup_l CSS_cst_cup_cup_r.

Proposition fCSS_inc_stm_l X C s f m : finite_CSS (X,C) -> pme C (m ~~ m) -> finite_CSS (X cup sg (s,(f,m)), C).
Proof. intros (?,(?,?)) ?; auto. Qed.

Proposition CSS_inc_stm2_l X C s f s' f' m : CSS (X,C) -> pme C (m ~~ m) -> CSS (X cup (sg (s,(f,m)) cup sg (s',(f',m))), C).
Proof. auto. Qed.

Proposition fCSS_inc_stm2_l X C s f s' f' m : finite_CSS (X,C) -> pme C (m ~~ m) -> finite_CSS (X cup (sg (s,(f,m)) cup sg (s',(f',m))), C).
Proof. intros (?,(?,?)) ?; auto. Qed.

Hint Resolve CSS_inc_stm_l CSS_inc_stm2_l fCSS_inc_stm2_l.

Proposition CSS_inc_r' X C D : C inc pme D -> CSS (X,C) -> CSS (X,D).
Proof.
  intros H1 H2 s f m H'.
  apply cl_inc in H1; auto.
  apply H1, H2 with (1 := H').
Qed.

Proposition CSS_inc_r X C D : C inc D -> CSS (X,C) -> CSS (X,D).
Proof. intros H; apply CSS_inc_r', inc_transitive with (1 := H), pme_inc; auto. Qed.

Hint Resolve CSS_inc_r' CSS_inc_r.

Proposition CSS_id_simpl X C X' C' : X ~ X' -> C ~ C' -> CSS (X,C) -> CSS (X',C').
Proof. intros ? ?; apply CSS_id_id; auto. Qed.

Proposition CSS_equiv_eps X C : CSS_equiv (X,C) (X,C cup sg (eps~~eps)).
Proof. apply CSS_equiv_sym; split; auto; apply pme_eps_1. Qed.

Proposition CSS_cup_rl X C1 C2 : CSS (X,C1) -> CSS (X, C1 cup C2).
Proof. intros; apply CSS_inc_r with (C := C1); auto. Qed.

Proposition CSS_cup_rr X C1 C2 : CSS (X,C2) -> CSS (X, C1 cup C2).
Proof. intros; apply CSS_inc_r with (C := C2); auto. Qed.

Hint Resolve CSS_cup_rl CSS_cup_rr.

Proposition fCSS_inc_r X C d : finite_CSS (X,C) -> finite_CSS (X,C cup sg d).
Proof. intros (?,(?,?)); auto. Qed. 

Proposition CSS_cupcup X1 X2 X3 C1 C2 C3 : 
                                 CSS (X1,C1) 
                              -> CSS (X2,C2) 
                              -> CSS (X3,C3)
                              -> CSS ((X1 cup X2) cup X3, (C1 cup C2) cup C3).
Proof. auto. Qed.

Hint Resolve CSS_cupcup.

Proposition fCSS_cupcup X1 X2 X3 C1 C2 C3 : 
                                 finite_CSS (X1,C1) 
                              -> finite_CSS (X2,C2) 
                              -> finite_CSS (X3,C3)
                              -> finite_CSS ((X1 cup X2) cup X3, (C1 cup C2) cup C3).
Proof. intros (?,(?,?)) (?,(?,?)) (?,(?,?)); auto. Qed.

Section SubstCSS.
  
  Variable sigma : label -> word label.
  
  Definition subst_stm (S : stm) : stm := match S with (b,(f,m)) => (b,(f,subst_w sigma m)) end.
  Definition subst_pstm := pd_image subst_stm.
  Definition subst_CSS p := match p with (X,C) => (subst_pstm X, subst_dimage sigma C) end.
  
  Proposition subst_pstm_mono (X : _ -> Prop) s f m : X (s,(f,m)) -> subst_pstm X (s,(f, subst_w sigma m)).
  Proof. intros HX; red; exists (s,(f,m)); auto. Qed.
  
  Proposition CSS_subst_stable : forall css, CSS css -> CSS (subst_CSS css).
  Proof.
    intros [] ? b f m [ [ b' [ f' m' ] ] [ H' Hs ] ].
    simpl in Hs.
    injection Hs. intros Hm _ _. rewrite <- Hm. 
    apply subst_pme_prop, (H b' f' m'); auto.
  Qed.
  
  Proposition subst_pstm_inc X X' : X inc X' -> subst_pstm X inc subst_pstm X'.
  Proof. intros HX; apply pd_image_inc; auto. Qed.
  
  Hint Resolve CSS_subst_stable subst_pstm_inc.
  
  Proposition CSS_subst_inc : forall css1 css2, CSS_leq css1 css2 -> CSS_leq (subst_CSS css1) (subst_CSS css2).
  Proof. intros [] [] (HX,HC). unfold subst_CSS. split; simpl; auto. apply subst_dimage_inc; auto. Qed.

  Hint Resolve CSS_subst_inc.

  Proposition CSS_subst_id : forall css1 css2, CSS_id css1 css2 -> CSS_id (subst_CSS css1) (subst_CSS css2).
  Proof. intros ? ? H; apply CSS_leq_antisym in H; apply CSS_leq_antisym; destruct H; auto. Qed.

  Proposition CSS_subst_finite : forall css, finite_CSS css -> finite_CSS (subst_CSS css).
  Proof.
    intros [] (?,(?,?)).
      split; auto.
      split; apply finite_pd_image; auto.
  Qed.

End SubstCSS.

Hint Resolve CSS_subst_stable subst_pstm_inc CSS_subst_inc CSS_subst_id CSS_subst_finite.

Section CSS_subst_alphabet.

  Variable X : stm -> Prop.
  Variable C : cst -> Prop.
  Variables s s' : label -> word label.
  Hypothesis Halpha : forall x, alphabet C x -> s x = s' x.
  
  Proposition subst_pstm_alpha_inc : CSS (X,C) -> subst_pstm s X inc subst_pstm s' X.
  Proof.
    intros HCSS. 
    intros (s0,(f0,m0)) ((s1,(f1,m1)),(H0,H1)). 
    simpl in H1; injection H1; clear H1; intros H1 H2 H3;
    rewrite <- H1, <- H2, <- H3; clear H1 H2 H3.
    exists (s1,(f1,m1)). 
    split; auto. 
    simpl. f_equal. f_equal.
    symmetry.
    apply subst_carrier_eq with (1 := Halpha). 
    apply pme_inc_alphabet.
    apply HCSS with (1 := H0).
  Qed.

  Proposition subst_dimage_alpha_inc : CSS (X,C) -> subst_dimage s C inc subst_dimage s' C.
  Proof.
    intros HCSS (m0,n0) ((m1,n1),(H0,H1)). 
    injection H1; clear H1; intros H1 H2;
    rewrite <- H1, <- H2; clear H1 H2.
    exists (m1,n1). 
    split; auto.
    simpl.
    
    f_equal; symmetry;
    apply subst_carrier_eq with (1 := Halpha);
    apply pme_inc_alphabet.
    
    apply pme_lft with (y := n1), pme_inc; auto.
    apply pme_rt with  (x := m1), pme_inc; auto.
  Qed.
  
End CSS_subst_alphabet.

Hint Resolve subst_pstm_alpha_inc subst_dimage_alpha_inc.

Section CSS_subst_alphabet'.
  
  Variable X : stm -> Prop.
  Variable C : cst -> Prop.
  Variables s s' : label -> word label.
  Hypothesis Halpha : forall x, alphabet C x -> s x = s' x.
  Hypothesis HXC : CSS (X,C).
  
  Proposition subst_pstm_alpha_id : subst_pstm s X ~ subst_pstm s' X.
  Proof. split; apply subst_pstm_alpha_inc with (C := C); auto. intros; symmetry; auto. Qed.
  
  Proposition subst_dimage_alpha_id : subst_dimage s C ~ subst_dimage s' C. 
  Proof. split; apply subst_dimage_alpha_inc with (X := X) (C := C); auto; intros; symmetry; auto. Qed.

  Hint Resolve subst_pstm_alpha_id subst_dimage_alpha_id.
  
  Proposition subst_CSS_alpha_id : CSS_id (subst_CSS s (X,C)) (subst_CSS s' (X,C)).
  Proof. simpl; split; auto. Qed.

End CSS_subst_alphabet'.

Hint Resolve subst_pstm_alpha_id subst_dimage_alpha_id subst_CSS_alpha_id.

Proposition subst_pstm_alpha_id' X : subst_pstm (fun x => l x) X ~ X.
Proof.
  apply pd_image_identity.
  intros (s,(f,m)).
  simpl; rewrite subst_w_id; auto.
Qed.

Proposition subst_dimage_alpha_id' C : subst_dimage (fun x : label => l x) C ~ C.
Proof.
  apply pd_image_identity.
  intros (m,n); simpl; rewrite subst_w_id, subst_w_id; auto.
Qed.

Hint Resolve subst_pstm_alpha_id' subst_dimage_alpha_id'.

Proposition subst_CSS_id' css : CSS_id (subst_CSS (fun x => l x) css) css.
Proof. destruct css; simpl; auto. Qed.


