(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import NPeano.
Require Import List.

Require Import predicate.
Require Import maps.
Require Import natseq.
Require Import enum.
Require Import nat2.
Require Import maxf.

Section NatRepeat.

  Definition nat_repeat n : nat := snd (idx_dmap _ nat2_index n).

  Let nr_inv j i := idx_rmap _ nat2_index (i,j).

  Let nr_inv_inj j : fnat_inj (nr_inv j).
  Proof.
    intros i i' H.
    assert ((i,j)=(i',j)) as H'.
      rewrite <- (idx_surj _ nat2_index (i,j)); trivial.
      rewrite <- (idx_surj _ nat2_index (i',j)); trivial.
      unfold nr_inv in H. rewrite -> H. reflexivity.
    injection H'; trivial.
  Qed.

  Proposition nat_repeat_isurj : i_surjective nat_repeat.
  Proof.
    intros x n.
    destruct (fnat_inj_prop _ (nr_inv_inj x) n) as [ k  [ Hk1 Hk2 ] ].
    exists (nr_inv x k). 
    split; auto.
    unfold r_image, nat_repeat, nr_inv.
    rewrite -> (idx_surj _ nat2_index); trivial.
  Qed.

End NatRepeat.

Definition enum_bool : enumeration bool.
Proof.
  set (dB := fun c => match c with O => false | _ => true end).
  set (rB := fun c => match c with false => O | true => 1 end).
  exists dB rB.
  intros c H. induction c; simpl; reflexivity.
Qed.

Definition nat_enum : enumeration nat.
Proof. apply (enum_from_idx _ nat_index). Qed.

Definition enum_prod X Y (eX : enumeration X) (eY : enumeration Y) : enumeration (X*Y).
Proof.
  destruct eX as [dX rX sX].
  destruct eY as [dY rY sY].
  destruct nat2_index as [dN rN _ sN ].
  set (dP := fun c => match dN c with (a,b) => (dX a,dY b) end).
  set (rP := fun c => match c with (x,y) => rN (rX x, rY y) end).
  exists dP rP.
  intros c H.
  destruct c as [x y].
  unfold dP. unfold rP. 
  rewrite -> sN; trivial.
  rewrite -> sX; trivial.
  rewrite -> sY; trivial.
Qed.

Definition enum_sum X Y (eX : enumeration X) (eY : enumeration Y) : enumeration (X+Y).
Proof.
  destruct eX as [dX rX sX].
  destruct eY as [dY rY sY].
  destruct nat2_index as [dN rN _ sN ].
  set (dP := fun c => match dN c with (0,a) => inl _ (dX a) | (_,b) => inr _ (dY b) end).
  set (rP := fun c => match c with inl x => rN (0, rX x) | inr y => rN (1, rY y) end).
  exists dP rP.
  intros c H.
  destruct c as [x | y]; unfold dP; unfold rP; rewrite -> sN; trivial. 
  rewrite -> sX; trivial. rewrite -> sY; trivial.
Qed.

Definition enum_union X (phi : nat -> X -> Prop) (psy : X -> nat) 
                   (t : forall x, (union phi) x -> phi (psy x) x) 
                   (e : forall n, partial_enumeration (phi n)) : partial_enumeration (union phi).
Proof.
  destruct nat2_index as [dm rm _ srj].
  set (dmap0 := fun n => match dm n with (a,b) => enum_dmap _ (e a) b end).
  set (rmap0 := fun x => rm (psy x, enum_rmap _ (e (psy x)) x)).
  exists dmap0 rmap0.
  intros; unfold dmap0, rmap0.
  rewrite -> srj; trivial.
  destruct (e (psy x)) as [dm1 sm1 srj1]. simpl.
  apply srj1. apply t. trivial.
Qed.

Definition enum_layers X (phi : nat -> X -> Prop) (psy : X -> nat) 
                                              (t : forall x, phi (psy x) x) 
                                              (e : forall n, partial_enumeration (phi n)) 
                                            : enumeration X.
Proof.
  apply (penum_id _ (union phi)).
  split; intros x Hx. trivial. exists (psy x). apply t.
  apply (enum_union _ _ psy); trivial.
Qed.

Section ListSurj.

  Variable A : Type.
  Variable eA : enumeration A.

  Let list_bounded n (l : list A) := length l <= n.

  Let list_enum_rec n : partial_enumeration (list_bounded n).
  Proof.
    destruct eA as [dA rA sA].
    induction n.
    
    set (dL := fun x : nat => (nil : list A)).
    set (rL := fun l : list A => O).
    exists dL rL.
    intros. unfold rL. unfold dL. induction x. reflexivity.
    red in H. simpl in H. contradiction (le_Sn_O _ H).

    destruct IHn as [dTn rTn sTn].
    set (eHn := enum_sum _ _ nat_enum (enum_prod _ _ nat_enum nat_enum)).
    destruct eHn as [dN rN sN].
    set (dL := fun n => match dN n with inl n => (nil : list A) 
                                                             | inr (a,b) => (dA a)::(dTn b) 
                                    end).
    set (rL := fun t => match t with nil => rN (inl _ 0) | a::b => rN (inr _ (rA a, rTn b)) end).
    exists dL rL.
    induction x as [ _ | a l _ ]; unfold dL; unfold rL; 
    rewrite -> sN; trivial.
    intros Hl.
    red in Hl. simpl in Hl.
    apply le_S_n in Hl.
    rewrite -> (sTn l); trivial.
    rewrite -> sA; trivial.
  Qed.
    
  Definition list_enum : enumeration (list A).
  Proof.
    apply (enum_layers _ list_bounded (length (A:=A)) ).
    intros; red; auto.
    apply list_enum_rec.
  Qed.

  Let a := enum_dmap _ eA 0.
  Let seq := let fix loop n { struct n } := match n with 0 => nil | S n => a::(loop n) end in loop.
  Let sze (l : list A) := length l.

  Definition list_index : indexation (list A).
  Proof.
    apply idx_from_enum_iseq.
    apply list_enum.
    exists seq sze.
    intros n; simpl; auto.
  Qed.

End ListSurj.

Section TreeSurj.

  Variable A : Type. 
  Variable eA : enumeration A.

  Inductive bin_tree : Type :=
    | leaf : A -> bin_tree
    | node : bin_tree -> bin_tree -> bin_tree.

  Let max_prop a b c : max a b <= c -> a<=c /\ b<= c.
  Proof. intros; split; apply le_trans with (m := max a b); auto with arith. Qed.

  Let bt_height : bin_tree -> nat := 
    let fix loop (t : bin_tree) { struct t } := match t with
      | leaf n => 0
      | node a b => S (max (loop a) (loop b))
    end in loop. 

  Let bt_bounded n t := bt_height t <= n.

  Let bt_enum_rec n : partial_enumeration (bt_bounded n).
  Proof.
    destruct eA as [dA rA sA].
    induction n.

    set (dT := fun n => leaf (dA n)).
    set (rT := fun t => match t with leaf a => rA a | _ => 0 end).
    exists dT rT. intros x Hx. induction x. 
    simpl. unfold dT. rewrite -> sA; trivial.
    red in Hx. simpl in Hx. contradiction (le_Sn_O _ Hx).

    destruct IHn as [dTn rTn sTn].
    set (eHn := enum_sum _ _ nat_enum (enum_prod _ _ nat_enum nat_enum)).
    destruct eHn as [dN rN sN].
    set (dT := fun n => match dN n with inl n => leaf (dA n) | inr (a,b) => node (dTn a) (dTn b) end).
    set (rT := fun t => match t with leaf a => rN (inl _ (rA a)) | node a b => rN (inr _ (rTn a,rTn b)) end).
    exists dT rT.
    induction x as [ | x1 _ x2 _ ]; intros Hx.
    unfold dT. simpl. rewrite -> sN; trivial. rewrite -> sA; trivial. 
    red in Hx. simpl in Hx. apply le_S_n in Hx.
    apply max_prop in Hx. destruct Hx as [Hx1 Hx2].
    unfold dT. simpl. rewrite -> sN; trivial.
    rewrite -> (sTn x1); trivial.
    rewrite -> (sTn x2); trivial.
  Qed.

  Definition bin_tree_enum : enumeration bin_tree.
  Proof.
    apply (enum_layers _ bt_bounded bt_height).
    intros; red; auto.
    apply bt_enum_rec.
  Qed.

  Let a := enum_dmap _ eA 0.
  Let seq := let fix loop n { struct n } := match n with 0 => leaf a | S n => node (leaf a) (loop n) end in loop.
  
  Definition bin_tree_index : indexation bin_tree.
  Proof.
    apply idx_from_enum_iseq.
    apply bin_tree_enum.
    exists seq bt_height.
    intros n; simpl; auto.
  Qed.

End TreeSurj.

Definition nat_bt_surj := enum_dmap _ (bin_tree_enum nat nat_enum).

Proposition bt_surj : surjective nat_bt_surj.
Proof. apply enumeration_surj. Qed.

Section DecidableTotalOrder.

  Variable X : Type.
  Variable eX : enumeration X.

  Let i := enum_rmap _ eX.

  Let inj : injective i.
  Proof. apply enumeration_inj. Qed.

  Definition dto_enum a b := (i a ) <= (i b).
  
  Proposition dto_enum_refl : forall x, dto_enum x x.
  Proof. intros. red; auto. Qed.

  Proposition dto_enum_antisym : forall x y, dto_enum x y -> dto_enum y x -> x = y.
  Proof. unfold dto_enum. intros. apply inj. apply le_antisym; auto. Qed.

  Proposition dto_enum_trans : forall x y z, dto_enum x y -> dto_enum y z -> dto_enum x z.
  Proof. unfold dto_enum; intros ? ? ?; apply le_trans. Qed.

  Proposition dto_enum_dec : forall x y,  { dto_enum x y } + { ~ dto_enum x y}.
  Proof. intros x y. unfold dto_enum. apply le_dec. Qed.

  Proposition dto_enum_total : forall x y, { dto_enum x y } + { dto_enum y x }.
  Proof. intros. unfold dto_enum. destruct (le_ge_dec (i x) (i y)); tauto. Qed.

End DecidableTotalOrder.

