(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Section list_fall_exst.
 
  Variables (X : Type) (P : X -> Prop).

  Proposition list_fall_cons x l : Forall P (x::l) -> P x /\ Forall P l.
  Proof. intros H; inversion_clear H; auto. Qed.

  Definition list_sig_fall (ll : list (sig P)) : Forall P (map (proj1_sig (P:=P)) ll).
  Proof.
    induction ll as [ | [x Hx] ll ]; constructor; simpl; auto.
  Defined.

  Let list_fall_sig_rec ll : Forall P ll -> { l : list (sig P) | ll = map (proj1_sig (P:=P)) l }.
  Proof.
    induction ll; intros H.
    exists nil; auto.
    apply list_fall_cons in H.
    destruct H as [ Ha Hll ].
    apply IHll in Hll.
    destruct Hll as [ l Hll ].
    exists ((exist _ _ Ha)::l).
    rewrite Hll.
    auto.
  Qed.

  Definition list_fall_sig ll Hll : list (sig P) := proj1_sig (list_fall_sig_rec ll Hll).

  Fact list_fall_sig_prop ll Hll : ll = map (proj1_sig (P:=P)) (list_fall_sig ll Hll).
  Proof. apply (proj2_sig (list_fall_sig_rec ll Hll)). Qed.
   
  Proposition list_nfall_cons x l : ~(~ P x \/ ~ Forall P l) -> ~ ~ Forall P (x::l).
  Proof.
    intros H0 H1.
    assert (~ (P x /\ Forall P l)) as H'.
    intros [H2 H3]; apply H1; constructor; auto.
    tauto.
  Qed.

  Proposition list_fall_app_0 l : Forall P l -> forall x, In x l -> P x.
  Proof. intros H; apply Forall_forall; auto. Qed.
  
  Proposition list_fall_app_1 (l r : list X) : Forall P (l++r) -> Forall P l /\ Forall P r.
  Proof.
    intros H; split;
    apply Forall_forall; intros x H';
    apply (proj1 (Forall_forall P (l++r)) H);
    apply in_or_app; [ left | right ]; auto.
  Qed.

  Proposition list_fall_app_2 (l r : list X) : Forall P l -> Forall P r -> Forall P (l++r).
  Proof.
    intros H1 H2.
    apply Forall_forall; intros x H'.
    apply in_app_or in H'.
    destruct H' as [ H' | H' ].
    apply (proj1 (Forall_forall P _) H1 _ H').
    apply (proj1 (Forall_forall P _) H2 _ H').
  Qed.

  Proposition list_nfall_app_1 (l r : list X) : ~ (~ Forall P l \/ ~ Forall P r) -> ~ ~ Forall P (l++r).
  Proof.
    intros H H'.
    assert (~ (Forall P l /\ Forall P r)) as H''.
    intros [ H1 H2 ]; apply H'; apply list_fall_app_2; auto.
    tauto.
  Qed.

  Proposition list_exst_lr x : P x -> forall l r, Exists P (l++x::r).
  Proof. induction l; intros r; simpl; auto. Qed.
  
  Proposition list_exst_r (r : list X) : Exists P r -> forall l, Exists P (l++r).
  Proof. induction l; simpl; auto. Qed.

  Proposition list_exst_l (l r: list X) : Exists P l -> Exists P (l++r).
  Proof. 
    intros; induction l; simpl; auto. 
    apply Exists_nil in H; contradiction H.
    apply Exists_cons in H; destruct H as [ H | H ].
    constructor; auto.
    apply Exists_cons_tl; auto.
  Qed.

  Proposition list_exst_2 ll : Exists P ll -> exists x, exists l, exists r, P x /\ ll = l++x::r.
  Proof.
    induction 1.
    exists x. exists nil. exists l. split; auto.
    destruct IHExists as [ y [ l0 [ l1 [ H1 H2 ] ] ] ].
    exists y. exists (x::l0). exists l1. split; auto.
    simpl. rewrite H2. auto.
  Qed.

End list_fall_exst.

Hint Resolve list_fall_app_2 list_exst_lr list_exst_l list_exst_r.

Section list_fall_exst_dec.

  Variables (X : Type) (P : X -> Prop).
  Hypothesis Pdec : forall x, { P x } + { ~ P x }.
  
  Proposition list_fall_exst_dec ll : { Forall P ll } + { Exists (fun x => ~ P x) ll }.
  Proof.
    induction ll.
    left; auto.
    destruct (Pdec a) as [ Ha | Ha ].
    destruct IHll as [ H | H ].
    left; auto.
    right; auto.
    right; auto.
  Qed.
  
  Proposition list_fall_dec ll : { Forall P ll } + { ~ Forall P ll }.
  Proof.
    destruct (list_fall_exst_dec ll) as [ H | H ].
    left; auto.
    right.
    intros H'.
    apply list_exst_2 in H.
    destruct H as [ x [ l [ r [ H1 H2 ] ] ] ].
    apply H1.
    rewrite H2 in H'.
    apply (proj1 (Forall_forall _ _) H').
    apply in_or_app. right. left. trivial.
  Qed.
    
End list_fall_exst_dec.

Section in_splitted.

  Variables (X : Type) (P : X -> Prop).

  Theorem in_splitted_1 ll (x:X) l1 l2 : ll=l1++x::l2 -> In x ll.
  Proof. intros H; rewrite H; apply in_or_app; right; left; auto. Qed.

  Theorem in_splitted_2 l1 l2 (x y: X) : In x (l1++y::l2) -> (x=y) \/ In x (l1++l2).
  Proof.
    intros H.
    apply in_app_or in H.
    destruct H. right; apply in_or_app; left; auto.
    apply in_inv in H.
    destruct H.
    left; auto.
    right; 
    apply in_or_app; right; auto.
  Qed.

  Theorem in_splitted_2' l1 l2 (x y : X) : ~ P x -> P y ->  In x (l1++y::l2) -> In x (l1++l2).
  Proof.
    intros H1 H2 H3.
    apply in_splitted_2 in H3.
    destruct H3 as [ H3 | ]; auto.
    rewrite H3 in H1.
    contradiction (H1 H2).
  Qed.

End in_splitted.

Ltac list_fall_simpl_one H := generalize (proj1(Forall_forall _ _) H); clear H; intro H.
Ltac list_fall_simpl := repeat
  match goal with
    | [H: Forall _ _ |- _ ] => list_fall_simpl_one H
  end.

Ltac list_fall_auto := repeat 
   match goal with
     | [H:   Forall _ nil    |- _     ] => clear H 
     | [H:   Forall _ (_::_) |- _     ] => apply list_fall_cons in H; destruct H
     | [H:   Forall _ (_++_) |- _     ] => apply list_fall_app_1 in H; destruct H
     | [H: ~ Forall _ (_::_) |- False ] => generalize H; clear H; apply list_nfall_cons; intros [ H | H ]
     | [H: ~ Forall _ (_++_) |- False ] => generalize H; clear H; apply list_nfall_app_1; intros [ H | H ]
   end.

Ltac tac_In_goal := auto;
  match goal with
    | [  |- In _ (_::_) ] => try (apply in_eq; fail); apply in_cons; tac_In_goal
    | [  |- In _ (_++_) ] => apply in_or_app; try (left; tac_In_goal; fail); right; tac_In_goal
  end.

Ltac tac_In_hyps := 
  match goal with 
    | [ H : In _ nil      |- _ ] => contradiction (in_nil H)
    | [ H : In _ (_::_) |- _ ] => apply in_inv in H; destruct H as [ H | H ]
    | [ H : In _ (_++_) |- _ ] => apply in_app_or in H; destruct H as [ H | H ]
  end.

Ltac In_auto := repeat tac_In_hyps; try tac_In_goal.


