(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import list_fall_exst.
Require Import list_in_t.

Require Import predicate.
Require Import closure.
Require Import maps.
Require Import carrier.
Require Import finite.

Require Import system.
Require Import system_extensions.

Require Import words.
Require Import constraints.

Section PME.

  Set Implicit Arguments.

  Variable letter : Type.

  Section Constraint_rules.

    Section rules_prems_concs.

      Definition c_rule_eps_c := (eps ~~ eps : constraint letter).

      Variable p : (word letter*word letter*word letter)%type.

      Let x := match p with (x,_,_) => x end.
      Let y := match p with (_,y,_) => y end.
      Let z := match p with (_,_,z) => z end.
      
      Definition c_rule_comm_p1 := x o y ~~ x o y.
      Definition c_rule_comm_c  := x o y ~~ y o x.
      
      Definition c_rule_sym_p1  := x ~~ y.
      Definition c_rule_sym_c   := y ~~ x.
      
      Definition c_rule_tran_p1 := x ~~ y.
      Definition c_rule_tran_p2 := y ~~ z.
      Definition c_rule_tran_c  := x ~~ z.
      
      Definition c_rule_sub_p1  := x o y ~~ x o y.
      Definition c_rule_sub_c   :=   x   ~~   x.
      
      Definition c_rule_comp_p1 := z o y ~~ z o y.
      Definition c_rule_comp_p2 :=   x   ~~   y.
      Definition c_rule_comp_c  := z o x ~~ z o y.

    End rules_prems_concs.
    
    Definition c_rule_eps  := MkUAxiom  c_rule_eps_c.
    Definition c_rule_comm := MkUSingle c_rule_comm_p1 c_rule_comm_c.
    Definition c_rule_sym  := MkUSingle c_rule_sym_p1 c_rule_sym_c.
    Definition c_rule_tran := MkUDouble c_rule_tran_p1 c_rule_tran_p2 c_rule_tran_c.
    Definition c_rule_sub  := MkUSingle c_rule_sub_p1 c_rule_sub_c.
    Definition c_rule_comp := MkUDouble c_rule_comp_p1 c_rule_comp_p2 c_rule_comp_c.
    
    Definition c_sys_PME := c_rule_eps::c_rule_comm::c_rule_sym::c_rule_tran::c_rule_sub::c_rule_comp::nil.
    
    Definition is_pme := closed_under_rules  c_sys_PME.
    Definition pme    := closure_under_proof c_sys_PME.

    (* these are generic properties that hold for ANY list of rules *)
    
    Proposition pme_is_pme R : is_pme (pme R).
    Proof. apply closure_under_proof_is_closed. Qed.
    
    Proposition is_pme_postfixpoint R : is_pme R <-> pme R inc R.
    Proof.
      split.
      apply closed_under_rules_postfixpoint.
      apply postfixpoint_closed_under_rules.
    Qed.
    
    Proposition pme_mono R S : R inc S -> pme R inc pme S.
    Proof. apply closure_under_proof_is_mono. Qed.

    Proposition pme_inc R : R inc pme R.
    Proof. apply closure_under_proof_is_inc. Qed.
    
    Proposition pme_idem R : pme (pme R) inc pme R. 
    Proof. apply closure_under_proof_is_idem. Qed.

    Proposition pme_mono' : Closure_monotonic pme.        Proof. exact pme_mono. Qed.
    Proposition pme_inc'  : Closure_increasing pme.       Proof. exact pme_inc. Qed.
    Proposition pme_idem' : Closure_idempotent pme.       Proof. exact pme_idem. Qed.
    
    Proposition pme_cpct : Closure_compact pme.
    Proof. intros C x H; destruct closure_under_proof_compactness with (1 := H) as ( Rf ); exists Rf; tauto. Qed.
    
    Proposition pme_min C meq : is_pme meq -> C inc meq -> pme C inc meq.
    Proof.
      intros H1 ?; apply is_pme_postfixpoint in H1.
      apply inc_transitive with (2 := H1), pme_mono; auto.
    Qed.  

    Proposition pme_cl_inc R S : R inc pme S -> pme R inc pme S.
    Proof. apply pme_min, pme_is_pme. Qed.

  End Constraint_rules.  

  Hint Resolve pme_is_pme is_pme_postfixpoint pme_mono pme_inc pme_idem pme_cpct pme_min.

  Proposition is_pme_fixpoint R : is_pme R -> R ~ pme R.
  Proof. split; auto. Qed.

  Proposition pme_compact Cf R : finite Cf -> Cf inc pme R -> exists Rf, finite Rf /\ Rf inc R /\ Cf inc pme Rf.
  Proof. apply cl_compact; auto. red; apply pme_mono. Qed.

  Proposition pme_id R S : R ~ S -> pme R ~ pme S.
  Proof. intros [ ]; split; auto. Qed.
  
  Proposition is_pme_id R S : R ~ S -> is_pme R -> is_pme S.
  Proof.
    intros [ H1 H2 ] H3.
    apply is_pme_fixpoint in H3.
    apply is_pme_postfixpoint.
    apply inc_transitive with (2 := H1),
          inc_transitive with (2 := proj2 H3); auto.
  Qed.
  
  Hint Resolve pme_id is_pme_id.
    
  Section is_pme_closed.

    Variable ceq : constraint letter -> Prop.
    Hypothesis ceqPME : is_pme ceq.
  
    Ltac solve r p := 
      intros;
      match goal with [ _: ceq ?x, _: ceq ?y |- _ ] => apply (@ceqPME r (x::y::nil))
                    | [ _: ceq ?x            |- _ ] => apply (@ceqPME r (x::nil))
                    | [                      |- _ ] => apply (@ceqPME r nil)
      end; unfold c_sys_PME; In_t_auto; exists p; simpl; auto.

    Proposition ceq_eps : ceq (eps ~~ eps).
    Proof. solve c_rule_eps tt. Qed.
    
    Proposition ceq_comm x y : ceq (x o y ~~ x o y) -> ceq (x o y ~~ y o x).
    Proof. solve c_rule_comm (x,y,y). Qed.
    
    Proposition ceq_sub x y : ceq (x o y ~~ x o y) -> ceq (x ~~ x).
    Proof. solve c_rule_sub (x,y,y). Qed.
    
    Proposition ceq_comp k x y : ceq (k o y ~~ k o y) -> ceq (x ~~ y) -> ceq (k o x ~~ k o y).
    Proof. solve c_rule_comp (x,y,k). Qed.
    
    Proposition ceq_tran x y z : ceq (x ~~ y) -> ceq (y ~~ z) -> ceq (x ~~ z).
    Proof. solve c_rule_tran (x,y,z). Qed.
    
    Proposition ceq_sym x y : ceq (x ~~ y) -> ceq (y ~~ x).
    Proof. solve c_rule_sym (x,y,y). Qed.
    
    (* hence pme are closed under the intented rules, as expected *)
    
    Hint Resolve   ceq_eps ceq_comm ceq_sub ceq_sym.
    
    Proposition ceq_lft x y : ceq (x ~~ y) -> ceq (x ~~ x).
    Proof. intros H; apply ceq_tran with (1 := H); auto. Qed.
    
    Proposition ceq_rt x y : ceq (x ~~ y) -> ceq (y ~~ y).
    Proof. intros H; apply ceq_tran with (2 := H); auto. Qed.
    
    Hint Resolve ceq_lft ceq_rt.
    
    Proposition ceq_comm_l x y m : ceq ( x o y ~~ m ) -> ceq (y o x ~~ m).
    Proof. intros H; apply ceq_tran with (2 := H), ceq_sym, ceq_comm, ceq_lft with (1 := H). Qed.
    
    Proposition ceq_comm_r x y m : ceq ( m ~~ x o y ) -> ceq (m ~~ y o x).
    Proof. intros H; apply ceq_sym in H; apply ceq_sym, ceq_comm_l; auto. Qed.
    
    Proposition ceq_pl k x y: ceq (k o x ~~ y) -> ceq (x ~~ x).
    Proof. intros H; apply ceq_lft, ceq_comm, ceq_rt, ceq_sub in H; auto. Qed.
    
    Proposition ceq_pr k x y : ceq (x ~~ k o y) -> ceq (y ~~ y).
    Proof. intros H; apply ceq_sym, ceq_pl in H; auto. Qed.
    
    Hint Resolve ceq_pl ceq_pr.
    
    Proposition ceq_el k x y m : ceq (x ~~ y) -> ceq (k o y ~~ m) -> ceq (k o x ~~ m).
    Proof.
      intros H1 H2; generalize (ceq_lft H2); intros H3.
      apply ceq_tran with (2 := H2), ceq_comp; auto.
    Qed.
    
    Proposition ceq_el' k x y m : ceq (x ~~ y) -> ceq (y o k ~~ m) -> ceq (x o k ~~ m).
    Proof.
      intros H1 H2; generalize (ceq_lft H2); intros H3.
      apply ceq_comm, ceq_sym in H3.
      generalize (ceq_el _ H1 (ceq_tran H3 H2)); intros H4.
      generalize (ceq_lft H4); intros H5.
      apply ceq_comm in H5.
      apply ceq_tran with (2 := H4); auto.
    Qed.
    
    Proposition ceq_weq_l x y m : x weq y -> ceq (x ~~ m) -> ceq (y ~~ m).
    Proof.
      intros H; generalize H m; clear m H.
      induction 1 as [ | a x x' | a b x | ]; intros m; auto.
      cutrewrite (a::x = l a o x); auto.
      cutrewrite (a::x' = l a o x'); auto.
      intros H1.
      generalize (IHPermutation _ (ceq_pl _ _ H1)); intros H2.
      apply ceq_el with (1 := H2); auto.
      intros H1.
      generalize (ceq_lft H1).
      intros H2.
      cutrewrite (a::b::x = (l a o l b) o x); auto.
      cutrewrite (b::a::x = (l b o l a) o x) in H2; auto.
      cutrewrite (b::a::x = (l b o l a) o x) in H1; auto.
      apply ceq_el' with (2 := H1).
      apply ceq_sym, ceq_comm.
      apply ceq_lft, ceq_sub in H1; auto.
    Qed.
    
    Proposition ceq_weq_r x y m : x weq y -> ceq (m ~~ x) -> ceq (m ~~ y).
    Proof.
      intros H1 H2; apply ceq_sym in H2.
      apply ceq_sym, ceq_weq_l with (1 := H1); auto.
    Qed.
    
    Hint Resolve ceq_weq_l ceq_weq_r.
    
    Proposition ceq_weq x1 x2 y1 y2 : x1 weq x2 -> y1 weq y2 -> ceq (x1 ~~ y1) -> ceq (x2 ~~ y2).
    Proof. intros; eauto. Qed.
    
    Proposition ceq_language_1 x : language ceq x -> ceq (x ~~ x).
    Proof.
      intros Hx.
      destruct Hx as ( a & b & Hab & [ [ k Hx ] | [ k Hx ] ] ).
      apply weq_sym in Hx.
      apply ceq_weq_l with (1 := Hx) in Hab.
      apply ceq_comm_l in Hab.
      apply ceq_pl in Hab; auto.
      apply weq_sym in Hx.
      apply ceq_weq_r with (1 := Hx) in Hab.
      apply ceq_comm_r in Hab.
      apply ceq_pr in Hab; auto.
    Qed.
    
    Proposition ceq_language_2 x : ceq (x ~~ x) -> language ceq x.
    Proof.
      intros Hx.
      exists x; exists x.
      split; auto.
      left; exists eps. 
      rewrite -> eps_neutral_r; auto.
    Qed.
  
    Proposition ceq_alphabet_1 a : alphabet ceq a -> ceq (l a ~~ l a).
    Proof. apply ceq_language_1. Qed.
    
    Proposition ceq_alphabet_2 a : ceq (l a ~~ l a) -> alphabet ceq a.
    Proof. apply ceq_language_2. Qed.
    
  End is_pme_closed.
  
  Section closed_is_pme.

  (* every relation closed under the following rules is a pme *)  
  
    Variable meq : constraint letter -> Prop.

    Hypothesis meq_eps  : meq (eps ~~ eps).
    Hypothesis meq_comm : forall x y, meq (x o y ~~ x o y) -> meq (x o y ~~ y o x).
    Hypothesis meq_sub  : forall x y, meq (x o y ~~ x o y) -> meq (x ~~ x).
    Hypothesis meq_comp : forall k x y, meq (k o y ~~ k o y) -> meq (x ~~ y) -> meq (k o x ~~ k o y).
    Hypothesis meq_tran : forall x y z, meq (x ~~ y) -> meq (y ~~ z) -> meq (x ~~ z).
    Hypothesis meq_sym  : forall x y, meq (x ~~ y) -> meq (y ~~ x).
    
    Proposition meq_is_pme : is_pme meq.
    Proof.
      intros r hh c Hr I HH.
      unfold c_sys_PME in Hr; In_t_auto; rewrite Hr in I; red in I; simpl in I;
      destruct I as [ x [ _ H ] ]; 
      injection  H; intros H1 H2; rewrite H1; rewrite H2 in HH; clear H1 H2 H;
      (try apply meq_eps); destruct x as [ [ x y ] z ]; list_fall_auto;
      repeat match goal with [ K: meq (?cr _) |- _ ] => unfold cr in K 
                            |[ |- meq (?cr _)      ] => unfold cr end; auto.
      apply meq_tran with (y := y); auto.
      apply meq_sub with (y := y); auto.
    Qed.
  
  End closed_is_pme.

  Section pme_props.
  
    Ltac solve x := apply x with (1 := @pme_is_pme _).
  
    Proposition pme_eps R         : pme R (eps ~~ eps).      Proof. solve ceq_eps. Qed.
    
    Proposition pme_comm R x y    : pme R (x o y ~~ x o y) 
                                 -> pme R (x o y ~~ y o x).  Proof. solve ceq_comm. Qed.
                                 
    Proposition pme_sub R x y     : pme R (x o y ~~ x o y) 
                                 -> pme R     (x ~~ x).      Proof. solve ceq_sub. Qed.

    Proposition pme_comp R k x y  : pme R (k o y ~~ k o y) 
                                 -> pme R     (x ~~ y)
                                 -> pme R (k o x ~~ k o y).  Proof. solve ceq_comp. Qed.
 
    Proposition pme_tran R x y z  : pme R (x ~~ y)
                                 -> pme R (y ~~ z) 
                                 -> pme R (x ~~ z).          Proof. solve ceq_tran. Qed.

    Proposition pme_sym R x y     : pme R (x ~~ y) 
                                 -> pme R (y ~~ x).          Proof. solve ceq_sym. Qed.
 
    Proposition pme_lft R x y     : pme R (x ~~ y) 
                                 -> pme R (x ~~ x).          Proof. solve ceq_lft. Qed.
   
    Proposition pme_rt R x y      : pme R (x ~~ y) 
                                 -> pme R (y ~~ y).          Proof. solve ceq_rt. Qed.

    Proposition pme_pl R k x y    : pme R (k o x ~~ y) 
                                 -> pme R     (x ~~ x).      Proof. solve ceq_pl. Qed.

    Proposition pme_pr R k x y    : pme R (x ~~ k o y) 
                                 -> pme R (y ~~ y).          Proof. solve ceq_pr. Qed.

    Proposition pme_el R k x y m  : pme R (x ~~ y) 
                                 -> pme R (k o y ~~ m) 
                                 -> pme R (k o x ~~ m).      Proof. solve ceq_el. Qed.
      
    Proposition pme_el' R k x y m : pme R (x ~~ y) 
                                 -> pme R (y o k ~~ m) 
                                 -> pme R (x o k ~~ m).      Proof. solve ceq_el'. Qed.

    Proposition pme_weq R x1 x2 y1 y2 : x1 weq x2 -> y1 weq y2 
                                     -> pme R (x1 ~~ y1) 
                                     -> pme R (x2 ~~ y2).    Proof. solve ceq_weq. Qed.

    Proposition language_pme C x : language (pme C) x <-> pme C (x ~~ x).
    Proof. split. solve ceq_language_1. apply ceq_language_2. Qed.
    
    Proposition pme_language C x : pme C (x ~~ x) -> language (pme C) x.
    Proof. apply ceq_language_2; auto. Qed.
    
    Proposition pme_alphabet C a : alphabet (pme C) a <-> pme C (l a ~~ l a).
    Proof. split. solve ceq_alphabet_1. apply ceq_alphabet_2. Qed.
    
  End pme_props.
  
  Hint Resolve pme_eps pme_comm pme_comp pme_sym.
  Hint Resolve pme_el pme_el' pme_weq.
  Hint Resolve language_pme pme_language pme_alphabet.
  
  Proposition pme_eps_1 R : pme (R cup sg (eps~~eps)) ~ pme R.
  Proof. split; auto. Qed.
  
  Hint Resolve pme_eps_1.

  Section pme_help.

    Variable C : constraint letter -> Prop.
    Variables x y z : word letter.
    
    Hypothesis HC : pme C (x o y ~~ z).
    
    Fact pme_help_lemma_1 : pme C (y ~~ y).
    Proof. apply pme_pl in HC; auto. Qed.
    
    Fact pme_help_lemma_2 : pme C (x ~~ x).
    Proof. apply pme_lft, pme_comm, pme_sym, pme_pl in HC; auto. Qed.
  
    Fact pme_help_lemma_4 : pme C (z ~~ z).
    Proof. apply pme_rt in HC; auto. Qed.

    Hint Resolve pme_help_lemma_1 pme_help_lemma_2 pme_help_lemma_4.
    
    Fact pme_help_lemma_9 : pme C (x ~~ x) /\ pme C (y ~~ y) /\ pme C (z ~~ z).
    Proof. auto. Qed.

  End pme_help.

  Fact pme_help_lemma_3 C x y m : pme C (x o m ~~ y) -> pme C (x ~~ x).
  Proof. apply pme_help_lemma_2. Qed.
  
  Fact pme_help_lemma_5 C1 C2 c : pme C2 c -> pme (C1 cup C2) c.
  Proof. apply pme_mono; auto. Qed.
  
  Fact pme_help_lemma_6 a m k : pme (sg ((a::m) ~~ k)) (l a ~~ l a).
  Proof.
    cutrewrite (a::m = l a o m); auto.
    apply pme_help_lemma_2 with (y := m) (z := k), pme_inc; auto.
  Qed.
  
  Fact pme_help_lemma_7 a m k : pme (sg ((a::m) ~~ k)) (m ~~ m).
  Proof.
    cutrewrite (a::m = l a o m); auto.
    apply pme_help_lemma_1 with (x := l a) (z := k), pme_inc; auto.
  Qed.

  Fact pme_help_lemma_8 m k : pme (sg (m ~~ k)) (k ~~ k).
  Proof. apply pme_rt with (x := m), pme_inc; auto. Qed.

  Section meq_L.
  
    Variable L : letter -> Prop.
  
    Definition meq_L (c : constraint letter) := match c with (m,n) => carrier m inc L /\ carrier n inc L end.
    
    Proposition R_inc_meq_L R : alphabet R inc L -> R inc meq_L.
    Proof.
      intros HR (x,y) Hxy.
      split; apply inc_transitive with (B := alphabet R); auto;
      intros a Ha; apply carrier_inc_subletter in Ha;
      exists x; exists y; tauto.
    Qed.
    
    Proposition meq_L_is_pme : is_pme meq_L.
    Proof.
      apply meq_is_pme; simpl; intros; pred_solve; 
      try (apply carrier_app_1 in H; destruct H; auto).
    Qed.
     
     Proposition meq_L_postfixpoint : pme meq_L inc meq_L.
     Proof. apply is_pme_postfixpoint, meq_L_is_pme. Qed.
     
     Proposition alphabet_meq_L_1 : alphabet meq_L inc L.
     Proof. intros ? (?,(?,([ ],[ Ha | Ha ]))); apply subletter_inc_carrier in Ha; auto. Qed.
     
     Proposition alphabet_meq_L_2 : L inc alphabet meq_L.
     Proof.
       intros a Ha. 
       assert (carrier (a::nil) inc L); auto.
       exists (l a); exists (l a). 
       split.
       simpl; auto.
       left; exists eps; simpl; auto.
     Qed.
     
     Hint Resolve alphabet_meq_L_1 alphabet_meq_L_2 .
     
     Proposition alphabet_meq_L : L ~ alphabet meq_L.
     Proof. split; auto. Qed.
     
     Proposition meq_L_o m n k : meq_L (m~~m) -> meq_L (n~~n) -> meq_L (k~~k) -> meq_L (m o n ~~ k).
     Proof. intros [] [] []; split; auto. Qed.
     
     Proposition meq_L_eps_eps : meq_L (eps~~eps).
     Proof. split; auto. Qed.
     
     Proposition meq_L_eps m : meq_L (m~~m) -> meq_L (eps~~m).
     Proof. intros []; split; auto. Qed.

  End meq_L.
  
  Hint Resolve R_inc_meq_L alphabet_meq_L alphabet_meq_L_1 alphabet_meq_L_2.
  
  Proposition meq_L_inc L L' : L inc L' -> meq_L L inc meq_L L'.
  Proof. intros ? [] []; split; apply inc_transitive with (B := L); auto. Qed.

  Hint Resolve meq_L_inc.
  
  Proposition alphabet_inc_meq_L R L : alphabet R inc L -> pme R inc meq_L L.
  Proof. intros; apply inc_transitive with (2 := proj1 (is_pme_postfixpoint _) (@meq_L_is_pme _)); auto. Qed.
  
  Hint Resolve alphabet_inc_meq_L.
  
  Proposition pme_inc_alphabet R m : pme R (m~~m) -> carrier m inc alphabet R.
  Proof.
    intros Hm.
    apply alphabet_inc_meq_L with (L := alphabet R) in Hm; auto.
    apply Hm.
  Qed. 
  
  Proposition alphabet_pme R : alphabet R ~ alphabet (pme R).
  Proof. split; auto; apply inc_transitive with (B := alphabet (meq_L (alphabet R))); auto. Qed.
  
  Proposition pme_empty : pme empty ~ sg (eps~~eps).
  Proof.
    split.
    intros (x,y) Hxy.
    apply (@alphabet_inc_meq_L empty empty) in Hxy.
    destruct Hxy as [ H0 H1 ].
    apply carrier_eps in H0. 
    apply carrier_eps in H1.
    rewrite H0, H1; auto.
    apply alphabet_empty.
    pred_solve.
  Qed.
  
  Proposition alphabet_pme_finite R : finite R -> finite (alphabet (pme R)).
  Proof. intro; apply finite_id with (1 := alphabet_pme _); auto. Qed.

End PME.

Hint Resolve pme_is_pme is_pme_postfixpoint.
Hint Resolve pme_mono pme_inc pme_idem.
Hint Resolve pme_mono' pme_inc' pme_idem'.
Hint Resolve pme_cpct pme_min pme_cl_inc.

Hint Extern 0 ( ?x inc pme ?x ) => apply pme_inc.
Hint Extern 1 ( pme _ inc pme _ ) => apply pme_mono.
Hint Extern 1 ( pme _ inc pme _ ) => apply pme_cl_inc. 
Hint Extern 1 ( _ inc pme _ ) => apply inc_transitive with (2 := pme_inc _).
Hint Extern 1 (pme _ _) => apply pme_inc.
	
Hint Resolve pme_id.
Hint Resolve pme_eps pme_comm pme_comp pme_sym.
Hint Resolve language_pme pme_language pme_alphabet.
Hint Resolve pme_eps_1.
Hint Resolve R_inc_meq_L alphabet_meq_L alphabet_meq_L_1 alphabet_meq_L_2.
Hint Resolve meq_L_inc.
Hint Resolve alphabet_inc_meq_L.
Hint Resolve pme_inc_alphabet alphabet_pme alphabet_pme_finite.
  
Section Substitutions.

  Set Implicit Arguments.

  Variable L M : Type.  

  Variable sigma : L -> word M.
  Let s := subst_w sigma.
  
  Definition subst_cst (c : constraint L) := match c with (m,n) => s m ~~ s n end.   
  Definition subst_rimage (C : _ -> Prop) c := match c with (m,n) => C (s m ~~ s n) end.
  Definition subst_dimage := pd_image subst_cst.
  
  Proposition subst_cst_rimage C : subst_rimage C ~ pr_image subst_cst C.
  Proof.
    split; intros (m,n) Hmn.
    exists (s m ~~ s n); tauto.
    destruct Hmn as [ (x,y) (H1,H2) ].
    rewrite <- H2 in H1; auto.
  Qed.
  
  Proposition subst_dimage_inc C C' : C inc C' -> subst_dimage C inc subst_dimage C'.
  Proof. intros ?; apply pd_image_inc; auto. Qed.
  
  Proposition subst_rimage_inc C C' : C inc C' -> subst_rimage C inc subst_rimage C'.
  Proof. 
    intros ?. 
    apply inc_transitive with (1 := proj1 (subst_cst_rimage _)).
    apply inc_transitive with (2 := proj2 (subst_cst_rimage _)).
    apply pr_image_inc; auto. 
  Qed.
  
  Proposition subst_rimage_is_pme C : is_pme C -> is_pme (subst_rimage C).
  Proof.
    intros HC.
    apply meq_is_pme; unfold subst_rimage.
    
    apply ceq_eps with (1 := HC).
    
    intros x y H; unfold s in H |- *. 
    rewrite subst_o in H.
    rewrite subst_o, subst_o. 
    apply ceq_comm with (1 := HC); auto.
    
    intros x y H; unfold s in H |- *. 
    rewrite subst_o in H.
    apply ceq_sub with (1 := HC) (2 := H); auto.
    
    intros k x y H1 H2; unfold s in H1 |- *. 
    rewrite subst_o in H1.
    rewrite subst_o, subst_o.
    apply ceq_comp with (1 := HC); auto.
    
    intros ? ? ?; apply ceq_tran; auto.
    
    intros ? ?; apply ceq_sym; auto.
  Qed.
  
  Variable C : constraint L -> Prop.
  
  Proposition subst_dimage_alphabet : alphabet (subst_dimage C) ~ subst_alphabet sigma (alphabet C).
  Proof.
    split.

    intros a ( x & y & [ (m,n) [ Hmn Hxy ] ] & Ha ).
    injection Hxy; intros Hn Hm; rewrite <- Hm, <- Hn in * |- *; clear x y Hxy Hm Hn.
    destruct Ha as [ Ha | Ha ]; apply subletter_inc_carrier in Ha; [ exists m | exists n ]; split; eauto.

    intros a Ha; apply subst_alphabet_prop in Ha. 
    destruct Ha as ( b & ( m & n & Hmn & [ Hb | Hb ] ) & Ha ); apply subletter_inc_carrier in Hb;
    exists (s m); exists (s n); split; 
    try ( exists (m,n); tauto ); [ left | right ];
    apply carrier_inc_subletter, subst_carrier with (2 := Ha); auto. 
  Qed.
  
  Proposition subst_dimage_pme_inc : subst_dimage (pme C) inc pme (subst_dimage C).
  Proof.
    set (R := subst_rimage (pme (subst_dimage C))).
    assert (is_pme R) as H0.  
      apply subst_rimage_is_pme; auto.
    assert (C inc subst_rimage (pme (subst_dimage C))) as H1. 
      apply inc_transitive with (2 := proj2 (subst_cst_rimage _)),
            inc_transitive with (1 := prd_image_inc _ _ subst_cst _), 
            pr_image_inc, pme_inc.
    apply is_pme_postfixpoint in H0.
    apply cl_inc_closed with (3 := H0) in H1; auto.
    intros (x,y) ((m,n),(Hmn,Hxy)).
    rewrite <- Hxy; apply (H1 _ Hmn).
  Qed.
  
  Proposition subst_pme_prop m n : pme C (m ~~ n) -> pme (subst_dimage C) (subst_w sigma m ~~ subst_w sigma n).
  Proof. intro; apply subst_dimage_pme_inc; exists (m,n); tauto. Qed.
  
End Substitutions.

Proposition subst_rimage_id {L} (C : _ -> Prop) m n : C (m ~~ n) -> subst_rimage (fun x : L => l x) C (m ~~ n).
Proof. intros; red; rewrite -> subst_w_id, -> subst_w_id; auto. Qed.

Proposition subst_dimage_id' {L} C : subst_dimage (fun x : L => l x) C ~ C.
Proof.
  intros. split; intros (x,y) H.
  
  destruct H as [ (x',y') (H1,H2) ].
  simpl in H2.
  rewrite subst_w_id, subst_w_id in H2. 
  rewrite <- H2;  trivial.
  
  exists (x,y). 
  split; auto. 
  simpl; rewrite subst_w_id, subst_w_id; auto.
Qed.

